from sklearn.tree import DecisionTreeRegressor

from biosphere import DecisionTree

from .common import Benchmark


class BiosphereTree(Benchmark):
    name = "biosphere tree"
    param_names = ["n", "mtry"]
    params = ([1000, 10000, 100000, 1000000], [4, 12])

    def _setup_model(self, params):
        _, max_features = params

        # inspect.getargspec(DecisionTree.__init__) does not work
        try:
            self.model = DecisionTree(max_depth=8, max_features=max_features,)
        # For biosphere<0.3.0, max_features was called mtry
        except TypeError:
            self.model = DecisionTree(max_depth=8, mtry=max_features)


class ScikitLearnTree(Benchmark):
    name = "scikit-learn tree"
    param_names = ["n", "mtry"]
    params = ([1000, 10000, 100000, 1000000], [4, 12])

    def _setup_model(self, params):
        _, mtry = params
        self.model = DecisionTreeRegressor(max_depth=8, max_features=mtry,)



# From https://github.com/xhochy/nyc-taxi-fare-prediction-deployment-example
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.compose import make_column_transformer
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer, OneHotEncoder

DATA_PATH = Path(__file__).resolve().parent / "data"
NYC_TAXI_DATASET_PATH = DATA_PATH / "nyc_taxi_data.parquet"
NYC_TAXI_DATASET_URL = (
    "https://s3.amazonaws.com/nyc-tlc/trip+data/yellow_tripdata_2019-01.csv"
)
NYC_TAXI_BURROWS_PATH = DATA_PATH / "nyc_taxi_burrows.parquet"
NYC_TAXI_BURROWS_URL = "https://s3.amazonaws.com/nyc-tlc/misc/taxi+_zone_lookup.csv"


def load_nyc_taxi_raw():
    if NYC_TAXI_DATASET_PATH.exists():
        return pd.read_parquet(NYC_TAXI_DATASET_PATH)
    else:
        df = pd.read_csv(
            NYC_TAXI_DATASET_URL,
            dtype={"store_and_fwd_flag": "bool"},
            parse_dates=["tpep_pickup_datetime", "tpep_dropoff_datetime"],
            index_col=False,
            infer_datetime_format=True,
            true_values=["Y"],
            false_values=["N"],
        )
        df.to_parquet(NYC_TAXI_DATASET_PATH)
        return df


def load_nyc_taxi_burrows():
    if NYC_TAXI_BURROWS_PATH.exists():
        return pd.read_parquet(NYC_TAXI_BURROWS_PATH)
    else:
        df = pd.read_csv(NYC_TAXI_BURROWS_URL, index_col=False,)
        df.to_parquet(NYC_TAXI_BURROWS_PATH)
        return df


def split_pickup_datetime(df):
    return pd.DataFrame(
        {
            "pickup_dayofweek": df["tpep_pickup_datetime"].dt.dayofweek,
            "pickup_hour": df["tpep_pickup_datetime"].dt.hour,
            "pickup_minute": df["tpep_pickup_datetime"].dt.minute,
        }
    )


def burrow(df):
    df = df.merge(
        load_nyc_taxi_burrows().rename(columns={"Borough": "pickup_borough"}),
        left_on="PULocationID",
        right_on="LocationID",
        how="left",
    )
    df["pickup_borough"] = df["pickup_borough"].fillna("Unknown")
    return df[["pickup_borough"]]


burrow_categories = [
    "Bronx",
    "Brooklyn",
    "EWR",
    "Manhattan",
    "Queens",
    "Staten Island",
    "Unknown",
]


def feature_enginering():
    return make_column_transformer(
        (FunctionTransformer(), ["passenger_count", "trip_distance"]),
        (FunctionTransformer(func=split_pickup_datetime), ["tpep_pickup_datetime"],),
        (
            Pipeline(
                steps=[
                    ("pickup_borough", FunctionTransformer(burrow)),
                    ("one_hot_encoder", OneHotEncoder(categories=[burrow_categories])),
                ]
            ),
            ["PULocationID"],
        ),
    )


def load_nyc_taxi(train_size, test_size):
    df = load_nyc_taxi_raw().head(train_size + test_size)
    df = df[lambda x: x["fare_amount"] > 0]
    df["fare_amount"] = np.log(df["fare_amount"])

    fare_amount_mean = df["fare_amount"].mean()
    fare_amount_std = df["fare_amount"].std()
    df = df[
        lambda x: x["fare_amount"].between(
            fare_amount_mean - 4 * fare_amount_std,
            fare_amount_mean + 4 * fare_amount_std,
        )
    ]

    y = df.pop("fare_amount").to_numpy()
    X = feature_enginering().fit_transform(df)

    return train_test_split(X, y, test_size=test_size / (train_size + test_size))



import timeit
from abc import ABC, abstractmethod
from typing import Any

from sklearn.metrics import mean_squared_error

from .load import load_nyc_taxi


class Benchmark(ABC):
    # Measure wall time instead of CPU usage
    timer = timeit.default_timer

    name = None

    params: Any
    param_names: Any

    def _time_predict(self, *args):
        self.model.predict(self.X_test)

    def time_fit(self, *args):
        self.model.fit(self.X_train, self.y_train)

    def setup(self, *params):
        self._setup_data(params)
        self._setup_model(params)

    def _setup_data(self, params):
        idx = self.param_names.index("n")
        n = params[idx]
        self.X_train, self.X_test, self.y_train, self.y_test = load_nyc_taxi(n, n)

    @abstractmethod  # Required for asv to skip.
    def _setup_model(self, params):
        pass

    def score(self):
        return mean_squared_error(self.model.predict(self.X_test), self.y_test,)



import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error

from biosphere import RandomForest

from .common import Benchmark


class ScikitLearnForest(Benchmark):
    def __init__(self, n_jobs=2):
        self.n_jobs = n_jobs  # GitHub CI has two nodes.

    name = "scikit-learn forest"
    param_names = ["n", "n_estimators", "max_features"]
    # params=([1000, 10000, 100000], [100, 400], [4, 12], [2])
    params = ([10000], [100], [4, 12])

    def _setup_model(self, params):
        _, n_estimators, max_features = params
        self.model = RandomForestRegressor(
            n_estimators=n_estimators,
            max_depth=8,
            max_features=max_features,
            n_jobs=self.n_jobs,
            oob_score=True,
        )

    def _time_fit_predict_oob(self):
        self.oob_score = True
        self.model.fit(self.X_train, self.y_train)
        return (1 - self.model.oob_score_) * np.mean(
            (self.y_train - self.y_train.mean()) ** 2
        )


class BiosphereForest(Benchmark):
    def __init__(self, n_jobs=2):
        self.n_jobs = n_jobs  # GitHub CI has two nodes.

    name = "biosphere forest"
    param_names = ["n", "n_estimators", "max_features"]
    params = ([1000, 10000, 100000], [100], [4, 12])

    def _setup_model(self, params):
        _, n_estimators, max_features = params

        # inspect.getargspec(RandomForest.__init__) does not work
        try:
            self.model = RandomForest(
                max_depth=8,
                n_jobs=self.n_jobs,
                max_features=max_features,
                n_estimators=n_estimators,
            )
        # For biosphere<0.3.0, max_features was called mtry and n_trees was called
        # n_estimators.
        except TypeError:
            try:
                self.model = RandomForest(
                    max_depth=8,
                    n_jobs=self.n_jobs,
                    mtry=max_features,
                    n_trees=n_estimators,
                )
            # For biosphere<0.2.0, no parallelization
            except TypeError:
                self.model = RandomForest(max_depth=8, mtry=max_features)

    def _time_fit_predict_oob(self):
        predictions = self.model.fit_predict_oob(self.X_train, self.y_train)
        return mean_squared_error(self.y_train, predictions)



# Benchmarks for the README
import time

import pandas as pd

from benchmarks.benchmark_forest import BiosphereForest, ScikitLearnForest

benchmark_parameters = [
    (1000, 400, 12),
    (2000, 400, 12),
    (4000, 400, 12),
    (8000, 400, 12),
    (16000, 400, 12),
    (32000, 400, 12),
    (64000, 400, 12),
    (128000, 400, 12),
    (256000, 400, 12),
    (512000, 400, 12),
    (1024000, 400, 12),
    (2048000, 400, 12),
]

models = [ScikitLearnForest, BiosphereForest]

n_samples = 10

if __name__ == "__main__":
    results = pd.DataFrame(
        columns=["model", "n", "mtry", "n_estimators", "time", "score", "oob_score"]
    )

    for parameters in benchmark_parameters:
        for model in models:
            m = model(n_jobs=4)
            m.setup(*parameters)

            for _ in range(n_samples):
                tic = time.perf_counter_ns()
                oob_score = m._time_fit_predict_oob()
                toc = time.perf_counter_ns()

                score = m.score()
                results.loc[len(results)] = [
                    m.name,
                    *parameters,
                    toc - tic,
                    score,
                    oob_score,
                ]

            print(
                f"{model.name} {parameters} "
                f"time={results.tail(n_samples)['time'].min()/1e9:.4f}"
                f" score={score:.4f} oob_score={oob_score:.4f}"
            )

    results.to_csv("results.csv")



from pathlib import Path

import numpy as np
import pytest

from biosphere import DecisionTree, RandomForest

_IRIS_FILE = "iris.csv"
_IRIS_PATH = Path(__file__).resolve().parents[2] / "testdata" / _IRIS_FILE


def test_forest():
    data = np.loadtxt(_IRIS_PATH, skiprows=1, delimiter=",", usecols=(0, 1, 2, 3, 4))
    X = data[:, 0:4]
    y = data[:, 4]

    random_forest = RandomForest()
    oob_predictions = random_forest.fit_predict_oob(X, y)
    predictions = random_forest.predict(X)

    oob_mse = np.mean((oob_predictions - y) ** 2)
    mse = np.mean((predictions - y) ** 2)

    assert oob_mse < 0.05
    assert mse < oob_mse / 2


def test_tree():
    data = np.loadtxt(_IRIS_PATH, skiprows=1, delimiter=",", usecols=(0, 1, 2, 3, 4))
    X = data[:, 0:4]
    y = data[:, 4]

    decision_tree = DecisionTree()
    decision_tree.fit(X, y)
    predictions = decision_tree.predict(X)

    mse = np.mean((predictions - y) ** 2)

    assert mse < 0.05


# TODO: Better test checking that supplying parameters had correct effect.
@pytest.mark.parametrize("max_features", [0.2, 3, "sqrt", None])
def test_max_features(max_features):
    _ = RandomForest(max_features=max_features)
    _ = DecisionTree(max_features=max_features)
